@@ -2,7 +2,7 @@
  * ns6: A security assessment tool for attack vectors based on
  *      ICMPv6 Neighbor Solicitation messages
  *
- * Copyright (C) 2009-2013 Fernando Gont
+ * Copyright (C) 2009-2015 Fernando Gont
  *
  * Programmed by Fernando Gont for SI6 Networks <http://www.si6networks.com>
  *
@@ -18,7 +18,6 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
  * 
  * Build with: make ns6
  * 
@@ -75,8 +74,8 @@ unsigned long		ul_res, ul_val;
     
 unsigned int		i, j, startrand, sources, nsources, targets, ntargets;
     
-u_int16_t			mask;
-u_int8_t			hoplimit;
+uint16_t			mask;
+uint8_t			hoplimit;
 
 struct ether_addr	linkaddr[MAX_SLLA_OPTION];
 unsigned int		nlinkaddr=0, linkaddrs;
@@ -100,17 +99,17 @@ unsigned int		hbhopthdrlen[MAX_HBH_OPT_HDR], m, pad;
 
 struct ip6_frag		fraghdr, *fh;
 struct ip6_hdr		*fipv6;
-unsigned char		fragh_f=0;
 unsigned char		fragbuffer[ETHER_HDR_LEN+MIN_IPV6_HLEN+MAX_IPV6_PAYLOAD];
 unsigned char		*fragpart, *fptr, *fptrend, *ptrend, *ptrhdr, *ptrhdrend;
 unsigned int		hdrlen, ndstopthdr=0, nhbhopthdr=0, ndstoptuhdr=0;
-unsigned int		nfrags, fragsize, max_packet_size;
+unsigned int		nfrags, fragsize;
 unsigned char		*prev_nh, *startoffragment;
 struct iface_data	idata;
 
 int main(int argc, char **argv){
 	extern char	*optarg;
 	int			r;
+	struct target_ipv6	targetipv6;
 
 	static struct option longopts[] = {
 		{"interface", required_argument, 0, 'i'},
@@ -131,7 +130,8 @@ int main(int argc, char **argv){
 		{"loop", no_argument, 0, 'l'},
 		{"sleep", no_argument, 0, 'z'},
 		{"verbose", no_argument, 0, 'v'},
-		{"help", no_argument, 0, 'h'}
+		{"help", no_argument, 0, 'h'},
+		{0, 0, 0,  0 }
 	};
 
 	char shortopts[]= "i:s:d:A:u:U:H:y:S:D:t:eE:F:T:lz:vh";
@@ -187,11 +187,23 @@ int main(int argc, char **argv){
 				break;
 	    
 			case 'd':	/* IPv6 Destination Address */
-				if( inet_pton(AF_INET6, optarg, &(idata.dstaddr)) <= 0){
-					puts("inet_pton(): address not valid");
-					exit(EXIT_FAILURE);
+				strncpy( targetipv6.name, optarg, NI_MAXHOST);
+				targetipv6.name[NI_MAXHOST-1]= 0;
+				targetipv6.flags= AI_CANONNAME;
+
+				if( (r=get_ipv6_target(&targetipv6)) != 0){
+
+					if(r < 0){
+						printf("Unknown Destination: %s\n", gai_strerror(targetipv6.res));
+					}
+					else{
+						puts("Unknown Destination: No IPv6 address found for specified destination");
+					}
+
+					exit(1);
 				}
-		
+
+				idata.dstaddr= targetipv6.ip6;
 				idata.dstaddr_f = 1;
 				break;
 
@@ -209,7 +221,7 @@ int main(int argc, char **argv){
 				}
 		
 				nfrags = (nfrags +7) & 0xfff8;
-				fragh_f= 1;
+				idata.fragh_f= 1;
 				break;
 
 	    case 'u':	/* Destinations Options Header */
@@ -309,7 +321,7 @@ int main(int argc, char **argv){
 
 		hdrlen= atoi(optarg);
 		
-		if(hdrlen <= 8){
+		if(hdrlen < 8){
 		    puts("Bad length in Hop-by-Hop Options Header");
 		    exit(EXIT_FAILURE);
 		}
@@ -507,13 +519,12 @@ int main(int argc, char **argv){
 		   The KAME implementation discards addresses in which the second highe-order 16 bits
 		   (srcaddr.s6_addr16[1] in our case) are not zero.
 		 */  
-		idata.srcaddr.s6_addr16[0]= htons(0xfe80); /* Link-local unicast prefix */
+		if ( inet_pton(AF_INET6, "fe80::", &(idata.srcaddr)) <= 0){
+			puts("inet_pton(): Error when converting address");
+			exit(EXIT_FAILURE);
+		}
 
-		for(i=1;i<4;i++)
-			idata.srcaddr.s6_addr16[i]=0x0000;	
-	    
-		for(i=4; i<8; i++)
-			idata.srcaddr.s6_addr16[i]=random();
+		randomize_ipv6_addr(&(idata.srcaddr), &(idata.srcaddr), 64);
 	}
 
 
@@ -522,11 +533,12 @@ int main(int argc, char **argv){
 	   select the random Source Addresses from the link-local unicast prefix (fe80::/64).
 	 */
 	if(floods_f && !idata.srcprefix_f){
-		idata.srcaddr.s6_addr16[0]= htons(0xfe80); /* Link-local unicast prefix */
+		if ( inet_pton(AF_INET6, "fe80::", &(idata.srcaddr)) <= 0){
+			puts("inet_pton(): Error when converting address");
+			exit(EXIT_FAILURE);
+		}
 
-		for(i=1;i<8;i++)
-			idata.srcaddr.s6_addr16[i]=0x0000;
-	
+		randomize_ipv6_addr(&(idata.srcaddr), &(idata.srcaddr), 64);
 		idata.srcpreflen=64;
 	}
 
@@ -535,11 +547,12 @@ int main(int argc, char **argv){
 	   select the random Target Addresses from the link-local unicast prefix (fe80::/64).
 	 */
 	if(floodt_f && !targetprefix_f){
-		targetaddr.s6_addr16[0]= htons(0xfe80); /* Link-local unicast prefix */
+		if ( inet_pton(AF_INET6, "fe80::", &(targetaddr)) <= 0){
+			puts("inet_pton(): Error when converting address");
+			exit(EXIT_FAILURE);
+		}
 
-		for(i=1;i<8;i++)
-			targetaddr.s6_addr16[i]=0x0000;
-	
+		randomize_ipv6_addr(&(targetaddr), &(targetaddr), 64);
 		targetpreflen=64;
 	}
 
@@ -554,8 +567,7 @@ int main(int argc, char **argv){
 	}
 
 	if(!idata.hsrcaddr_f && !floods_f)	/* Source link-layer address is randomized by default */
-		for(i=0; i<6; i++)
-			idata.hsrcaddr.a[i]= random();
+		randomize_ether_addr(&(idata.hsrcaddr));
 
 	if(sllopt_f && !sllopta_f){			/* The value of the source link-layer address option  */
 		linkaddr[0]= idata.hsrcaddr;			/* defaults to the source Ethernet address            */
@@ -574,15 +586,10 @@ int main(int argc, char **argv){
 	if(!sleep_f)
 		nsleep=1;
 
-	if( !fragh_f && dstoptuhdr_f){
+	if( !idata.fragh_f && dstoptuhdr_f){
 		puts("Dst. Options Header (Unfragmentable Part) set, but Fragmentation not specified");
 		exit(EXIT_FAILURE);
 	}
-    
-	if(fragh_f)
-		max_packet_size = MAX_IPV6_PAYLOAD + MIN_IPV6_HLEN;
-	else
-		max_packet_size = ETH_DATA_LEN;
 
 	if(idata.verbose_f){
 		print_attack_info(&idata);
@@ -617,129 +624,142 @@ int main(int argc, char **argv){
  * that are expected to remain constant for the specified attack.
  */
 void init_packet_data(struct iface_data *idata){
-    ethernet= (struct ether_header *) buffer;
-    v6buffer = buffer + idata->linkhsize;
-    ipv6 = (struct ip6_hdr *) v6buffer;
+	struct dlt_null *dlt_null;
+	ethernet= (struct ether_header *) buffer;
+	dlt_null= (struct dlt_null *) buffer;
+	v6buffer = buffer + idata->linkhsize;
+	ipv6 = (struct ip6_hdr *) v6buffer;
 
-	if(idata->flags != IFACE_TUNNEL && idata->flags != IFACE_LOOPBACK){
-		ethernet->src = idata->hsrcaddr;
-		ethernet->dst = idata->hdstaddr;
+	if(idata->type == DLT_EN10MB){
 		ethernet->ether_type = htons(ETHERTYPE_IPV6);
+
+		if(!(idata->flags & IFACE_LOOPBACK)){
+			ethernet->src = idata->hsrcaddr;
+			ethernet->dst = idata->hdstaddr;
+		}
 	}
+	else if(idata->type == DLT_NULL){
+		dlt_null->family= PF_INET6;
+	}
+#if defined (__OpenBSD__)
+	else if(idata->type == DLT_LOOP){
+		dlt_null->family= htonl(PF_INET6);
+	}
+#endif
 
-    ipv6->ip6_flow=0;
-    ipv6->ip6_vfc= 0x60;
-    ipv6->ip6_hlim= hoplimit;
-    ipv6->ip6_src= idata->srcaddr;
-    ipv6->ip6_dst= idata->dstaddr;
-    prev_nh = (unsigned char *) &(ipv6->ip6_nxt);
+	ipv6->ip6_flow=0;
+	ipv6->ip6_vfc= 0x60;
+	ipv6->ip6_hlim= hoplimit;
+	ipv6->ip6_src= idata->srcaddr;
+	ipv6->ip6_dst= idata->dstaddr;
+	prev_nh = (unsigned char *) &(ipv6->ip6_nxt);
 
-    ptr = (unsigned char *) v6buffer + MIN_IPV6_HLEN;
+	ptr = (unsigned char *) v6buffer + MIN_IPV6_HLEN;
     
-    if(hbhopthdr_f){
-	hbhopthdrs=0;
+	if(hbhopthdr_f){
+		hbhopthdrs=0;
 	
-	while(hbhopthdrs < nhbhopthdr){
-	    if((ptr+ hbhopthdrlen[hbhopthdrs]) > (v6buffer+ ETH_DATA_LEN)){
-		puts("Packet too large while processing HBH Opt. Header");
-		exit(EXIT_FAILURE);
-	    }
+		while(hbhopthdrs < nhbhopthdr){
+			if((ptr+ hbhopthdrlen[hbhopthdrs]) > (v6buffer+ idata->mtu)){
+				puts("Packet too large while processing HBH Opt. Header");
+				exit(EXIT_FAILURE);
+			}
 	    
-	    *prev_nh = IPPROTO_HOPOPTS;
-	    prev_nh = ptr;
-	    memcpy(ptr, hbhopthdr[hbhopthdrs], hbhopthdrlen[hbhopthdrs]);
-	    ptr = ptr + hbhopthdrlen[hbhopthdrs];
-	    hbhopthdrs++;
+			*prev_nh = IPPROTO_HOPOPTS;
+			prev_nh = ptr;
+			memcpy(ptr, hbhopthdr[hbhopthdrs], hbhopthdrlen[hbhopthdrs]);
+			ptr = ptr + hbhopthdrlen[hbhopthdrs];
+			hbhopthdrs++;
+		}
 	}
-    }
 
-    if(dstoptuhdr_f){
-	dstoptuhdrs=0;
+	if(dstoptuhdr_f){
+		dstoptuhdrs=0;
 	
-	while(dstoptuhdrs < ndstoptuhdr){
-	    if((ptr+ dstoptuhdrlen[dstoptuhdrs]) > (v6buffer+ ETH_DATA_LEN)){
-		puts("Packet too large while processing Dest. Opt. Header (Unfrag. Part)");
-		exit(EXIT_FAILURE);
-	    }
+		while(dstoptuhdrs < ndstoptuhdr){
+			if((ptr+ dstoptuhdrlen[dstoptuhdrs]) > (v6buffer+ idata->mtu)){
+				puts("Packet too large while processing Dest. Opt. Header (Unfrag. Part)");
+				exit(EXIT_FAILURE);
+			}
 
-	    *prev_nh = IPPROTO_DSTOPTS;
-	    prev_nh = ptr;
-	    memcpy(ptr, dstoptuhdr[dstoptuhdrs], dstoptuhdrlen[dstoptuhdrs]);
-	    ptr = ptr + dstoptuhdrlen[dstoptuhdrs];
-	    dstoptuhdrs++;
+			*prev_nh = IPPROTO_DSTOPTS;
+			prev_nh = ptr;
+			memcpy(ptr, dstoptuhdr[dstoptuhdrs], dstoptuhdrlen[dstoptuhdrs]);
+			ptr = ptr + dstoptuhdrlen[dstoptuhdrs];
+			dstoptuhdrs++;
+		}
 	}
-    }
 
-    /* Everything that follows is the Fragmentable Part of the packet */
-    fragpart = ptr;
+	/* Everything that follows is the Fragmentable Part of the packet */
+	fragpart = ptr;
 
-    if(fragh_f){
-    	/* Check that we are able to send the Unfragmentable Part, together with a 
-    	   Fragment Header and a chunk data over our link layer
-    	 */
-    	if( (fragpart+sizeof(fraghdr)+nfrags) > (v6buffer+ETH_DATA_LEN)){
-    		puts("Unfragmentable part too large for current MTU (1500 bytes)");
-    		exit(EXIT_FAILURE);
-    	}
+	if(idata->fragh_f){
+		/* Check that we are able to send the Unfragmentable Part, together with a 
+		   Fragment Header and a chunk data over our link layer
+		 */
+		if( (fragpart+sizeof(fraghdr)+nfrags) > (v6buffer+idata->mtu)){
+			printf("Unfragmentable part too large for current MTU (%u bytes)\n", idata->mtu);
+			exit(EXIT_FAILURE);
+		}
 
-    	/* We prepare a separete Fragment Header, but we do not include it in the packet to be sent.
-    	   This Fragment Header will be used (an assembled with the rest of the packet by the 
-    	   send_packet() function.
-    	*/
-    	bzero(&fraghdr, FRAG_HDR_SIZE);
-    	*prev_nh = IPPROTO_FRAGMENT;
-    	prev_nh = (unsigned char *) &fraghdr;
-    }
+		/* We prepare a separete Fragment Header, but we do not include it in the packet to be sent.
+		   This Fragment Header will be used (an assembled with the rest of the packet by the 
+		   send_packet() function.
+		*/
+		memset(&fraghdr, 0, FRAG_HDR_SIZE);
+		*prev_nh = IPPROTO_FRAGMENT;
+		prev_nh = (unsigned char *) &fraghdr;
+	}
 
-    if(dstopthdr_f){
-	dstopthdrs=0;
+	if(dstopthdr_f){
+		dstopthdrs=0;
 	
-	while(dstopthdrs < ndstopthdr){
-	    if((ptr+ dstopthdrlen[dstopthdrs]) > (v6buffer+max_packet_size)){
-		puts("Packet too large while processing Dest. Opt. Header (U. part) (should be using the Frag. option?)");
-		exit(EXIT_FAILURE);
-	    }
+		while(dstopthdrs < ndstopthdr){
+			if((ptr+ dstopthdrlen[dstopthdrs]) > (v6buffer+idata->max_packet_size)){
+				puts("Packet too large while processing Dest. Opt. Header (U. part) (should be using the Frag. option?)");
+				exit(EXIT_FAILURE);
+			}
 	    
-	    *prev_nh = IPPROTO_DSTOPTS;
-	    prev_nh = ptr;
-	    memcpy(ptr, dstopthdr[dstopthdrs], dstopthdrlen[dstopthdrs]);
-	    ptr = ptr + dstopthdrlen[dstopthdrs];
-	    dstopthdrs++;
+			*prev_nh = IPPROTO_DSTOPTS;
+			prev_nh = ptr;
+			memcpy(ptr, dstopthdr[dstopthdrs], dstopthdrlen[dstopthdrs]);
+			ptr = ptr + dstopthdrlen[dstopthdrs];
+			dstopthdrs++;
+		}
 	}
-    }
 
-    *prev_nh = IPPROTO_ICMPV6;
+	*prev_nh = IPPROTO_ICMPV6;
 
-    if( (ptr+sizeof(struct nd_neighbor_solicit)) > (v6buffer+max_packet_size)){
-    	puts("Packet too large while inserting Neighbor Solicitation header (should be using Frag. option?)");
-    	exit(EXIT_FAILURE);
-    }
+	if( (ptr+sizeof(struct nd_neighbor_solicit)) > (v6buffer+idata->max_packet_size)){
+		puts("Packet too large while inserting Neighbor Solicitation header (should be using Frag. option?)");
+		exit(EXIT_FAILURE);
+	}
 
-    ns= (struct nd_neighbor_solicit *) (ptr);
+	ns= (struct nd_neighbor_solicit *) (ptr);
 
-    ns->nd_ns_type = ND_NEIGHBOR_SOLICIT;
-    ns->nd_ns_code = 0;
-    ns->nd_ns_reserved = 0;
-    ns->nd_ns_target = targetaddr;
+	ns->nd_ns_type = ND_NEIGHBOR_SOLICIT;
+	ns->nd_ns_code = 0;
+	ns->nd_ns_reserved = 0;
+	ns->nd_ns_target = targetaddr;
 
-    ptr += sizeof(struct nd_neighbor_solicit);
-    sllaopt = (struct nd_opt_slla *) ptr;    
+	ptr += sizeof(struct nd_neighbor_solicit);
+	sllaopt = (struct nd_opt_slla *) ptr;    
 
-    /* If a single source link-layer address is specified, it is included in all packets */
-    if(sllopt_f && nlinkaddr==1){
-    	if( (ptr+sizeof(struct nd_opt_slla)) <= (v6buffer+max_packet_size)){
-        	sllaopt->type= ND_OPT_SOURCE_LINKADDR;
-        	sllaopt->length= SLLA_OPT_LEN;
-        	bcopy(linkaddr[0].a, sllaopt->address, ETH_ALEN);
-        	ptr += sizeof(struct nd_opt_slla);
-        }
-        else{
-        	puts("Packet too large while processing source link-layer addresss opt. (should be using Frag. option?)");
-        	exit(EXIT_FAILURE);
-        }
-    }
+	/* If a single source link-layer address is specified, it is included in all packets */
+	if(sllopt_f && nlinkaddr==1){
+		if( (ptr+sizeof(struct nd_opt_slla)) <= (v6buffer+idata->max_packet_size)){
+			sllaopt->type= ND_OPT_SOURCE_LINKADDR;
+			sllaopt->length= SLLA_OPT_LEN;
+			memcpy(sllaopt->address, linkaddr[0].a, ETH_ALEN);
+			ptr += sizeof(struct nd_opt_slla);
+		}
+		else{
+			puts("Packet too large while processing source link-layer addresss opt. (should be using Frag. option?)");
+			exit(EXIT_FAILURE);
+		}
+	}
 
-    startofprefixes = ptr;
+	startofprefixes = ptr;
 }
 
 
@@ -759,38 +779,18 @@ void send_packet(struct iface_data *idata){
 			/* 
 			   Randomize the IPv6 Source address based on the specified prefix and prefix length
 			   (defaults to fe80::/64).
-			*/  
-			startrand= idata->srcpreflen/16;
-
-			for(i=0; i<startrand; i++)
-				ipv6->ip6_src.s6_addr16[i]= 0;
-
-			for(i=startrand; i<8; i++)
-				ipv6->ip6_src.s6_addr16[i]=random();
-
-
-			if(idata->srcpreflen%16){
-				mask=0xffff;
-	    
-				for(i=0; i<(idata->srcpreflen%16); i++)
-					mask= mask>>1;
-
-				ipv6->ip6_src.s6_addr16[startrand]= ipv6->ip6_src.s6_addr16[startrand] & htons(mask);
-			}
-
-			for(i=0; i<=(idata->srcpreflen/16); i++)
-				ipv6->ip6_src.s6_addr16[i]= ipv6->ip6_src.s6_addr16[i] | idata->srcaddr.s6_addr16[i];
+			*/
+			randomize_ipv6_addr(&(ipv6->ip6_src), &(idata->srcaddr), idata->srcpreflen);
 
 			if(!idata->hsrcaddr_f){
-				for(i=0; i<6; i++)
-					ethernet->src.a[i]= random();
+				randomize_ether_addr(&(ethernet->src));
 
 				/*
 				   If the source-link layer address must be included, but no value was 
 				   specified we set it to the randomized Ethernet Source Address
 				 */
 				if(sllopt_f && !sllopta_f){
-					bcopy(ethernet->src.a, sllaopt->address, ETH_ALEN);
+					memcpy(sllaopt->address, ethernet->src.a, ETH_ALEN);
 				}
 			}
 		}
@@ -803,28 +803,7 @@ void send_packet(struct iface_data *idata){
 				   Randomizing the ND Target Address based on the prefix specified by "targetaddr" 
 				   and targetpreflen.
 				 */  
-				startrand= targetpreflen/16;
-
-				for(i=0; i<startrand; i++)
-					ns->nd_ns_target.s6_addr16[i]= 0;
-
-				for(i=startrand; i<8; i++)
-					ns->nd_ns_target.s6_addr16[i]=random();
-
-				if(targetpreflen%16){
-					mask=0xffff;
-
-					for(i=0; i<(targetpreflen%16); i++)
-						mask= mask>>1;
-
-					ns->nd_ns_target.s6_addr16[startrand]= ns->nd_ns_target.s6_addr16[startrand] \
-													& htons(mask);
-				}
-
-				for(i=0; i<=(targetpreflen/16); i++)
-					ns->nd_ns_target.s6_addr16[i]= ns->nd_ns_target.s6_addr16[i] | \
-										targetaddr.s6_addr16[i];
-
+				randomize_ipv6_addr(&(ns->nd_ns_target), &(targetaddr), targetpreflen);
 			}
 
 			if(nlinkaddr==1)      /* If a single source link-layer address must be included, it is included */
@@ -836,11 +815,11 @@ void send_packet(struct iface_data *idata){
 				newdata_f=0;
 				ptr = startofprefixes;
 
-				while(linkaddrs<nlinkaddr && (ptr+sizeof(struct nd_opt_slla)-v6buffer)<=max_packet_size){
+				while(linkaddrs<nlinkaddr && (ptr+sizeof(struct nd_opt_slla)-v6buffer)<=idata->max_packet_size){
 					sllaopt = (struct nd_opt_slla *) ptr;
 					sllaopt->type= ND_OPT_SOURCE_LINKADDR;
 					sllaopt->length= SLLA_OPT_LEN;
-					bcopy(linkaddr[linkaddrs].a, sllaopt->address, ETH_ALEN);
+					memcpy(sllaopt->address, linkaddr[linkaddrs].a, ETH_ALEN);
 					ptr += sizeof(struct nd_opt_slla);
 					linkaddrs++;
 					newdata_f=1;
@@ -849,7 +828,7 @@ void send_packet(struct iface_data *idata){
 				ns->nd_ns_cksum = 0;
 				ns->nd_ns_cksum = in_chksum(v6buffer, ns, ptr-((unsigned char *)ns), IPPROTO_ICMPV6);
 
-				if(!fragh_f){
+				if(!idata->fragh_f){
 					ipv6->ip6_plen = htons((ptr - v6buffer) - MIN_IPV6_HLEN);
 
 					if((nw=pcap_inject(idata->pfd, buffer, ptr - buffer)) == -1){
@@ -1057,7 +1036,7 @@ void print_attack_info(struct iface_data *idata){
     for(i=0; i<ndstopthdr; i++)
 	printf("Destination Options Header: %u bytes\n", dstopthdrlen[i]);
 
-    if(fragh_f)
+    if(idata->fragh_f)
 	printf("Sending each packet in fragments of %u bytes (plus the Unfragmentable part)\n", nfrags);
 		
     if(!floodt_f){
